#ifdef MSWIN
	#define _CRT_RAND_S
#endif
#include <stdlib.h>
#include "VidHeader.cpp"
#include <math.h>
#include <stdio.h>
#include <limits.h>
#ifdef MSWIN
	#include <windows.h>
	typedef HANDLE THREAD_HANDLE;
#else
	#include <pthread.h>
	typedef pthread_t THREAD_HANDLE;
#endif

class ThreadData{
public:
	off_t	starti, endi;
	uchar	* uncompress_buffer;
	void init_data( off_t in_starti, off_t in_endi, uchar * in_uncompress_buffer ){
		starti				= in_starti;
		endi				= in_endi;
		uncompress_buffer	= in_uncompress_buffer;
	}
	void set_random_values(){
		off_t bufi = 0;
		while( bufi < endi ){
			off_t num_repeats	= (rand() * 20) / RAND_MAX;
			num_repeats			= my_min( num_repeats, endi - bufi );
			uchar rand_char		= (rand() * 30) / RAND_MAX;
			for( int repi = 0; repi < (int)num_repeats; repi++ )
				uncompress_buffer[bufi++]	= rand_char;
		}
	}
};

#ifdef MSWIN
DWORD WINAPI set_random_values_thread( LPVOID thread_arg ){
#else
void * set_random_values_thread( void * thread_arg ){
#endif
	ThreadData * thread_data 	= (ThreadData *)thread_arg;
	thread_data	-> set_random_values();
	return NULL;
}

int main( int argc, char ** argv ){
	
	ulong	compress_buffer_size	= 0;
	off_t	uncompress_buffer_size	= 0;
	uchar	* compress_buffer = NULL, * uncompress_buffer = NULL;
	
	#ifdef MSWIN
		MEMORYSTATUS memstat;
		GlobalMemoryStatus(&memstat);
		off_t	sys_memory			= memstat.dwTotalPhys;
		if( sys_memory > 6000000000 ){
			compress_buffer_size	= UINT_MAX/2;
		} else {
			compress_buffer_size	= (ulong)(sys_memory/4);
		}
		uncompress_buffer_size	= sys_memory - compress_buffer_size - UINT_MAX/7;
	#endif

	compress_buffer				= (uchar *) malloc(compress_buffer_size);
	uncompress_buffer			= (uchar *) malloc(uncompress_buffer_size);
	if( uncompress_buffer == NULL || compress_buffer == NULL ){
		fprintf(stderr, "Unable to allocate enough memory, this test program needs a system with at least 1Gb RAM.");
		exit(1);
	}

	#ifdef MSWIN
		fprintf( stderr, "Starting zlib based memory test using %I64d bytes.\n", compress_buffer_size + uncompress_buffer_size );
	#endif
	
	const int num_outer_iter			= 500;
	const int num_sample_iter			= 20;
	int num_all_iter					= num_sample_iter * num_outer_iter;
	int all_iter						= 1;
	
	int num_threads				= get_num_processing_cores();
	ThreadData * thread_data	= new ThreadData[num_threads];
	off_t block_size			= my_ceil( uncompress_buffer_size, num_threads );
	for( int i = 0; i < num_threads; i++ )
		thread_data[i].init_data( block_size*i, my_min( block_size*(i+1), uncompress_buffer_size), uncompress_buffer );

	THREAD_HANDLE * threads = new THREAD_HANDLE[num_threads];

	for( int outer_iter = 0; outer_iter < num_outer_iter; outer_iter++ ){
		srand( 12345  + 13*outer_iter );

		for( int i = 0; i < num_threads; i++ ){
			#ifdef MSWIN
				threads[i] = CreateThread( NULL, 0, set_random_values_thread, (LPVOID)(thread_data + i), 0, NULL );
			#else
				pthread_create( &threads[i], NULL, set_random_values_thread, (void *)(thread_data + i) );
			#endif
		}
		#ifdef MSWIN
			WaitForMultipleObjects(num_threads, threads, TRUE, INFINITE);
			for( int threadid = 0; threadid < num_threads; threadid++ )
				CloseHandle( threads[threadid] );
		#else
			for( int threadid = 0; threadid < num_threads; threadid++ )
				pthread_join( threads[threadid], NULL );
		#endif
		
		for( int samplei = 0;	samplei < num_sample_iter; samplei++, all_iter++ ){
			int	  first_try				= 1;		
			off_t  uncompress_offset	= 0;
			while( uncompress_offset + compress_buffer_size > uncompress_buffer_size || first_try ){
				uint rand_offset;	rand_s( &rand_offset );
				uncompress_offset	= rand_offset;
				uncompress_offset	*= my_ceil( (uncompress_buffer_size - compress_buffer_size), UINT_MAX );
				first_try			= 0;
			}
			#ifdef MSWIN
				fprintf( stderr, "Offset %I64d bytes.\n", uncompress_offset );
			#endif
			uchar * uncompress_buffer_sample	= uncompress_buffer + uncompress_offset;
			ulong compress_size					= compress_buffer_size;
			if( compress_MT( compress_buffer, compress_size, uncompress_buffer_sample, compress_buffer_size ) != Z_OK ){
				fprintf(stderr, "Compression error.\n"); exit(1);
			}
			ulong uncompress_size				= compress_buffer_size;
			if( uncompress_MT( uncompress_buffer_sample, uncompress_size, compress_buffer, compress_size ) != Z_OK ){
				fprintf(stderr, "De-compression error.\n"); exit(1);
			}
			//fprintf( stderr, "%10u, %10u, %10u \n", uncompress_offset, uncompress_size, compress_buffer_size );
			fprintf( stderr, "%6d of %6d \n", all_iter, num_all_iter );
		}
	}
	delete [] thread_data;
	delete [] threads;
}
/*
	ulong	compress_buffer_size	= UINT_MAX/2;
	off_t	uncompress_buffer_size	= (off_t)UINT_MAX*3 - (off_t)(compress_buffer_size) - off_t(UINT_MAX/4);
	
	while( compress_buffer_size > 0	&&  uncompress_buffer_size > 0 ){
		compress_buffer				= (uchar *) malloc(compress_buffer_size);
		uncompress_buffer			= (uchar *) malloc(uncompress_buffer_size);
		if( uncompress_buffer		!= NULL		&& compress_buffer != NULL )
			break;
		else{
			free( compress_buffer );
			free( uncompress_buffer );
		}
		compress_buffer_size		-= UINT_MAX/16;
		uncompress_buffer_size		-= UINT_MAX/8;
	}
	if( compress_buffer_size < 0	||  uncompress_buffer_size < 0 ){
		fprintf(stderr, "Unable to allocate enough memory, this test program needs a system with at least 1Gb RAM.");
		exit(1);
	}
*/